#include <Zmq/Zmq.mqh>

#define MONITOR_ADDR_RPC_SERVER  "inproc://monitor-rpc-server"

class ZmqServer {
  Context context_;
  Socket sock_;
  Socket monitor_;
  PollItem polls_[1];
  string endpoint_;
  bool running_;
  ZmqMsg zmsg_;

public:

  void ZmqServer():sock_(context_, ZMQ_REP), running_(false), monitor_(context_, ZMQ_PAIR) {
    sock_.monitor(MONITOR_ADDR_RPC_SERVER, ZMQ_EVENT_ALL);
    monitor_.connect(MONITOR_ADDR_RPC_SERVER);

    // sock_.fillPollItem(polls_[0], ZMQ_POLLIN);
    monitor_.fillPollItem(polls_[0], ZMQ_POLLIN);
  }

  bool Start(string endpoint) {
    endpoint_ = endpoint;
    running_  = sock_.bind(endpoint);
    if (running_)
      PrintFormat("Server started at %s", endpoint_);
    else
      PrintFormat("Server started failed %s", endpoint_);
    return running_;
  }

  void Stop(string endpoint) {
    running_ = false;
    monitor_.disconnect(MONITOR_ADDR_RPC_SERVER);
    sock_.disconnect(endpoint);
  }

  void Update() {
    if (!running_) return;

    Socket::poll(polls_, 0);

    // if (polls_[0].hasInput())
      recieve_zmq_message();

    if (polls_[0].hasInput())
      dispatch_monitor_events();

  }

  void HandleMessage(const uchar &req[], uchar &rep[]) {
    StringToCharArray("OK", rep);
  }

private:

  void recieve_zmq_message()
  {
    bool ok = sock_.recv(zmsg_, true);
    if (!ok) return;

    uchar req[];
    zmsg_.getData(req);

    //read more message body
    while (zmsg_.more())
    {
      zmsg_.rebuild();
      ok = sock_.recv(zmsg_, true);
      if (!ok) break;

      uchar buff[];
      zmsg_.getData(buff);
      ArrayResize(req, ArraySize(req) + ArraySize(buff));
      ArrayCopy(req, buff, ArraySize(req));
    }

    //5. handle message
    PrintFormat("[ZmqServer] recv msg: %s", CharArrayToString(req));
    uchar rep[];
    HandleMessage(req, rep);

    zmsg_.rebuild(ArraySize(rep));
    zmsg_.setData(rep);
    sock_.send(zmsg_);
    zmsg_.rebuild();
  }

  void dispatch_monitor_events()
  {
    ZmqMsg msg;
    if (monitor_.recv(msg))
    {
      if (msg.more())
      {
        //extract event
        uchar data[];
        msg.getData(data);
        int event = ((int)data[1] << 8) | ((int)data[0]);

        PrintFormat("ZmqServer got  monitor event: %d", event);
        //dispatch event
        switch (event)
        {
        case ZMQ_EVENT_CONNECTED:
          // OnConnected();
          break;
        }

        //get endpoint from msg more.
        zmq_msg_init(msg);
        monitor_.recv(msg, true);
      }
    }
  }

};

/////////////////////////////////////////////////////////////////////////////////
// EA
//////////////////////////////////////////////////////////////////////////////////

input string server_endpoint_ = "tcp://*:80";

ZmqServer server;

int OnInit() {

  server.Start(server_endpoint_);

  EventSetMillisecondTimer(10);
  return INIT_SUCCEEDED;
}

void OnTimer() {
  server.Update();
}

void OnTick() {
}

void OnDeinit(const int reason) {
  Print("OnDeinit()....");
  server.Stop(server_endpoint_);
}